import random

from typing import List

from ai_base import AItemplate, AIPlayer

import misc
from constants import *
import pygame
import random
import numpy as np
from collections import deque


# possible optimalization - if pheromones collide -> combine them
# (necessary to make pheromones more complex, but reduces number of sprites significantly)
class Pheromone(pygame.sprite.Sprite):
    def __init__(self, position, action, p_type, strength, lifespan):
        """

        :param position: tuple (x_pos, y_pos)
        """
        super(Pheromone, self).__init__()
        self.surf = pygame.Surface((5, 5))
        if p_type % 3 == 0:
            self.surf.fill((100 + 10 * p_type, 50 + 10 * p_type, 50 + 10 * p_type))
        elif p_type % 3 == 1:
            self.surf.fill((50 + 10 * p_type, 100 + 10 * p_type, 50 + 10 * p_type))
        else:
            self.surf.fill((50 + 10 * p_type, 50 + 10 * p_type, 100 + 10 * p_type))

        self.rect = self.surf.get_rect(center=position)

        self.lifespan = lifespan
        self.action = action
        self.strength = strength
        self.type = p_type


class Ant(AIPlayer):
    def __init__(self, spawnpoint):
        super(Ant, self).__init__(spawnpoint)
        self.past_positions = []
        self.past_actions = []
        self.touched_points = []

    def next_step(self, platforms, spec_points, pheromones=None):
        if pheromones is None:
            pheromones: List[Pheromone] = []
        super(Ant, self).next_step(platforms, spec_points)

        action_weights = [1, 1, 1, 1, 1, 1]

        # update probabilities of each action
        touched_pheromones = pygame.sprite.spritecollide(self, pheromones, dokill=False)
        for pheromone in touched_pheromones:
            if pheromone.type == self.spec_idx:
                action_weights[pheromone.action] += pheromone.strength

        # higher chance to make same action as last time
        if self.past_actions:
            action_weights[self.past_actions[-1]] *= 1.3

        # choose action
        cur_action = random.choices(range(len(action_weights)), weights=action_weights, k=1)[0]

        # going LEFT -> RIGHT and RIGHT -> LEFT is not allowed as it wastes a lot of moves
        # this just lowers the chance of it happening (sometimes it is correct to do so)
        if (cur_action == 1 and self.past_actions and self.past_actions[-1] == 5) or \
                (cur_action == 5 and self.past_actions and self.past_actions[-1] == 1):
            cur_action = random.choices(range(len(action_weights)), weights=action_weights, k=1)[0]

        pressed = misc.decode_action(cur_action)
        if cur_action >= 6:
            assert False, "invalid action"

        # remember position and chosen action (if not in finish yet)
        if self.spec_idx < len(spec_points):
            self.past_positions.append(self.rect.center)
            self.past_actions.append(cur_action)
        else:
            return True
        self.update(pressed, platforms, spec_points)
        return False

    def update(self, pressed_keys, platforms, spec_points=None):
        # 0 - nothing, 1 - left, 2 - top+left, 3 - top, 4 - top+right, 5 - right
        super(Ant, self).update(pressed_keys, platforms, spec_points)

        # touched next point of interest
        if self.spec_idx != len(self.touched_points):
            self.touched_points.append(len(self.past_actions))

    def reset_self(self):
        super(Ant, self).reset_self()
        self.past_positions = []
        self.past_actions = []
        self.touched_points = []


class AntColonyOrigin(AItemplate):
    name = 'AntColonyOrigin'

    def __init__(self, spawnpoint, spec_points, platforms, all_sprites):
        super().__init__(spawnpoint, spec_points, platforms, all_sprites)
        self.population = 20
        self.steps = 15
        self.step_inc = 5
        self.cur_step = 0
        self.pheromones = pygame.sprite.Group()
        self.pheromone_strength = 10.  # initial pheromone strength
        self.pheromone_evaporation = 0.7  # multiply pheromone strength after each step
        self.pheromone_lifetime = 3  # start removing pheromones after n steps
        # self.pheromone_destroyer_countdown = self.pheromone_lifetime

        self.pheromone_queue: deque[Pheromone] = deque()
        self.pheromone_dying: deque[int] = deque(
            [0 for _ in range(self.pheromone_lifetime)])  # how many items to remove from pheromone_queue
        self.colony: List[Ant] = []

        self.best_ant_ever = None
        self.ever_found_path = False
        self.best_found_path = None

    def prepare(self, use_default=False):
        # misc.get_numeric_input(default, min, max, text_info, num_type)
        if not use_default:
            self.population = misc.get_numeric_input(self.population, 1, 100, f"choose size of population", 'int')
            self.steps = misc.get_numeric_input(self.steps, 1, 1000000, f"choose starting number of steps", 'int')
            self.step_inc = misc.get_numeric_input(self.step_inc, 1, 1000000,
                                                   f"choose steps increase after each iteration",
                                                   'int')
            self.pheromone_strength = misc.get_numeric_input(self.pheromone_strength, 1, 100,
                                                             f"choose pheromone strength",
                                                             'float')
            self.pheromone_evaporation = misc.get_numeric_input(self.pheromone_evaporation, 0, 1,
                                                                f"choose pheromone evaporation rate", 'float')
            self.pheromone_lifetime = misc.get_numeric_input(self.pheromone_lifetime, 1, 100,
                                                             f"choose pheromone lifetime",
                                                             'int')

        # self.pheromone_destroyer_countdown = self.pheromone_lifetime
        self.pheromone_dying: deque[int] = deque([0 for _ in range(self.pheromone_lifetime)])

        for i in range(self.population):
            new_ant = Ant(self.spawnpoint)
            self.all_sprites.add(new_ant)
            self.colony.append(new_ant)

    def next_step(self):
        if self.cur_step >= self.steps:
            self.done = True
            return

        for ant in self.colony:
            if ant.next_step(self.platforms, self.spec_points, self.pheromones):
                self.done = True
                # print("found finish")
                self.found_path = True
                self.path = ant.past_actions[:self.cur_step]
        self.cur_step += 1

    def update_after_iteration(self):
        # print(f" {len(self.pheromone_dying)}")
        # print(f" pheromone queue len:{len(self.pheromone_queue)}")
        champion = self.evaluate_ants()
        # print(len(champion.past_actions))
        self.update_pheromones(champion)
        self.cur_step = 0
        self.steps += self.step_inc
        self.done = False
        self.found_path = False

    def reset_positions(self):
        for ant in self.colony:
            ant.reset_self()
            ant.rect.center = self.spawnpoint

    def update_best(self, champion):
        self.ever_found_path = champion.spec_idx >= len(self.spec_points)
        self.best_ant_ever = Ant(self.spawnpoint)
        self.best_ant_ever.touched_points = champion.touched_points
        self.best_ant_ever.past_actions = champion.past_actions
        self.best_ant_ever.past_positions = champion.past_positions
        self.best_ant_ever.spec_idx = champion.spec_idx
        self.best_found_path = champion.past_actions

    def evaluate_ants(self):
        # returns best ant
        best = 0
        best_ant = Ant(self.spawnpoint)
        for ant in self.colony:
            if not self.found_path:
                if ant.spec_idx == len(self.spec_points):
                    ant_result = ant.spec_idx + 1 / len(ant.past_positions)
                else:
                    ant_result = ant.spec_idx + 1 / misc.distance(ant.past_positions[-1],
                                                                  self.spec_points[ant.spec_idx].rect.center)
            else:
                ant_result = ant.spec_idx + 1 / len(ant.past_positions)
            if ant_result > best:
                best = ant_result
                best_ant = ant

        if not self.best_ant_ever:
            self.update_best(best_ant)
        else:
            if self.found_path and not self.ever_found_path:
                self.update_best(best_ant)
            else:
                if not self.ever_found_path:
                    former_best = self.best_ant_ever.spec_idx + 1 / misc.distance(self.best_ant_ever.past_positions[-1],
                                                                                  self.spec_points[
                                                                                      self.best_ant_ever.spec_idx].rect.center)
                else:
                    former_best = self.best_ant_ever.spec_idx + 1 / len(self.best_ant_ever.past_positions)

                if best > former_best:
                    self.update_best(best_ant)
                else:
                    return self.best_ant_ever
        return best_ant

    def update_pheromones(self, champion_ant: Ant):
        # evaluate fitness + update pheromones

        num = self.pheromone_dying.pop()
        # print(f" removing {num} pheromones")
        for i in range(num):
            self.pheromone_queue.pop().kill()

        for pheromone in self.pheromone_queue:
            pheromone.strength *= self.pheromone_evaporation

        # add new pheromones
        assert len(champion_ant.past_actions) == len(champion_ant.past_positions)
        self.pheromone_dying.appendleft(len(champion_ant.past_actions))
        # print(f" adding {len(champion_ant.past_actions)} new pheromones")
        pheromone_type = 0

        # weaken pheromones in case there is already some path
        pheromone_strength = self.pheromone_strength
        if self.found_path:
            pheromone_strength = self.pheromone_strength / 2.

        for i in range(len(champion_ant.past_actions)):
            if i in champion_ant.touched_points:
                pheromone_type += 1
            new_pheromone = Pheromone(champion_ant.past_positions[i], champion_ant.past_actions[i], pheromone_type,
                                      pheromone_strength, self.pheromone_lifetime)
            self.pheromones.add(new_pheromone)
            self.all_sprites.add(new_pheromone)
            self.pheromone_queue.appendleft(new_pheromone)
