import random

from typing import List

from ai_base import AItemplate, AIPlayer

import misc
from constants import *


class EvolPerson(AIPlayer):
    def __init__(self, spawnpoint, plan: list):
        super(EvolPerson, self).__init__(spawnpoint)
        self.spec_idx = 0
        self.current_step = 0
        self.player_plan = plan  # contains pressed keys for each frame
        self.fitness = 0
        self.finish_step = -1

    def next_step(self, platforms, spec_points):
        super(EvolPerson, self).next_step(platforms, spec_points)

        cur_action = self.player_plan[self.current_step]
        pressed = []
        if cur_action == 0:
            pass
        elif cur_action == 1:
            pressed = [K_LEFT]
        elif cur_action == 2:
            pressed = [K_LEFT, K_UP]
        elif cur_action == 3:
            pressed = [K_UP]
        elif cur_action == 4:
            pressed = [K_UP, K_RIGHT]
        elif cur_action == 5:
            pressed = [K_RIGHT]
        else:
            print("bug alert")

        self.current_step += 1
        if self.finish_step > 0:
            return True
        self.update(pressed, platforms, spec_points)
        return False

    def update(self, pressed_keys, platforms, spec_points=None):
        # 0 - nothing, 1 - left, 2 - top+left, 3 - top, 4 - top+right, 5 - right
        super(EvolPerson, self).update(pressed_keys, platforms, spec_points)
        if self.spec_idx >= len(spec_points):
            if self.finish_step < 0:
                self.finish_step = self.current_step
            return

    def reset_self(self):
        super(EvolPerson, self).reset_self()
        self.current_step = 0
        self.finish_step = -1


# class taking care of all EvolPerson instances
# ###################################### keep some survivors from old generation
class EvolAI(AItemplate):
    name = 'EvolAI'

    def __init__(self, spawnpoint, spec_points, platforms, all_sprites):
        super().__init__(spawnpoint, spec_points, platforms, all_sprites)
        self.population = 20
        self.steps = 15
        self.step_inc = 5
        self.cur_step = 0
        self.survivors = 3
        self.mut_rate = 0.25
        self.tournament_size = 3
        self.current_generation: List[EvolPerson] = []

        self.best_path = None
        self.ever_found_path = False
        self.best_found_path = None

    def prepare(self, use_default=False):
        # misc.get_numeric_input(default, min, max, text_info, num_type)
        if not use_default:
            self.population = misc.get_numeric_input(self.population, 1, 100, f"choose size of population", 'int')
            self.survivors = misc.get_numeric_input(self.survivors, 0, self.population-1, f"choose number of survivors", 'int')
            self.mut_rate = misc.get_numeric_input(self.mut_rate, 0., 1., f"choose mutation rate (interval 0-1)", 'float')
            self.steps = misc.get_numeric_input(self.steps, 1, 1000000, f"choose starting number of steps", 'int')
            self.step_inc = misc.get_numeric_input(self.step_inc, 1, 1000000, f"choose steps increase after each iteration", 'int')
            self.tournament_size = misc.get_numeric_input(self.tournament_size, 2, self.population-1, f"choose tournament size for choosing parents of next generation", 'int')

        self.generate_initial_generation()

    def generate_initial_generation(self):
        # 0 - nothing, 1 - left, 2 - right, 3 - top, 4 - top+left, 5 - top+right
        for i in range(self.population):
            new_individual = EvolPerson(self.spawnpoint, [random.randint(0, 5) for _ in range(self.steps)])
            self.all_sprites.add(new_individual)
            self.current_generation.append(new_individual)

    def evaluate_fitness(self):
        for individual in self.current_generation:
            # got to finish, shorter path is better
            if individual.spec_idx >= len(self.spec_points):
                individual.fitness = individual.spec_idx + 0.01 * (self.steps - individual.finish_step)
            else:
                individual.fitness = individual.spec_idx + 1 / misc.distance(individual.rect.center, self.spec_points[individual.spec_idx].rect.center)

    def select_parents(self, tournament_size=3):
        """Selects two parents from the population using tournament selection."""
        tournament = random.sample(self.current_generation, tournament_size)
        winner = max(tournament, key=lambda x: x.fitness)
        tournament.remove(winner)
        runner_up = max(tournament, key=lambda x: x.fitness)
        return winner, runner_up

    def create_next_generation(self):
        """Creates the next generation of the population."""
        new_population = []
        # select survivors from old generation
        survivors = sorted(self.current_generation, key=lambda x: x.fitness, reverse=True)[:self.survivors]
        for survivor in survivors:
            # elongate plan by adding random actions
            survivor.player_plan += [random.randint(0, 5) for _ in range(self.step_inc)]
            # reset survivors
            survivor.rect.center = self.spawnpoint
            survivor.current_step = 0
        new_population += survivors

        while len(new_population) < len(self.current_generation):
            parent1, parent2 = self.select_parents()
            child = self.crossover(parent1, parent2)
            if random.random() < self.mut_rate:
                child = mutate(child)
            # elongate plan by adding random actions
            child.player_plan += [random.randint(0, 5) for _ in range(self.step_inc)]
            new_population.append(child)
        return new_population

    # change it so that crossover_point is at checkpoints
    def crossover(self, parent1: EvolPerson, parent2: EvolPerson):
        """Performs crossover between two parents."""
        crossover_point = random.randint(1, len(parent1.player_plan) - 1)
        child = EvolPerson(self.spawnpoint,
                           plan=parent1.player_plan[:crossover_point] + parent2.player_plan[crossover_point:])
        return child

    def next_step(self):
        if self.cur_step >= self.steps:
            self.done = True
            return

        for p in self.current_generation:
            if p.next_step(self.platforms, self.spec_points):
                self.done = True
                self.ever_found_path = True
                self.found_path = True
                self.path = p.player_plan[:self.cur_step]
                if not self.best_found_path or len(self.best_found_path) > len(self.path):
                    self.best_found_path = self.path

        self.cur_step += 1

    def update_after_iteration(self):
        self.evaluate_fitness()
        next_generation = self.create_next_generation()
        # kill old generation
        for old in self.current_generation:
            old.kill()
        # ############################################ check if all new generation has same plan lengts
        for champ in next_generation:
            self.all_sprites.add(champ)
        self.current_generation = next_generation

        self.cur_step = 0
        self.steps += self.step_inc
        self.done = False
        self.found_path = False

    def reset_positions(self):
        for individual in self.current_generation:
            individual.reset_self()
            individual.rect.center = self.spawnpoint


def mutate(individual: EvolPerson):
    """Mutates an individual by changing a random action."""
    i = random.randint(0, len(individual.player_plan)-1)
    change = random.choice([1, -1])
    individual.player_plan[i] = (individual.player_plan[i] + change) % 6
    return individual

