from typing import List

import pygame
from matplotlib import pyplot as plt
from pygame.locals import *
import sys
import random
import numpy as np
import misc

HEIGHT = 1000
WIDTH = 400

FPS = 1200000

FramePerSec = pygame.time.Clock()

displaysurface = pygame.display.set_mode((WIDTH, HEIGHT))
pygame.display.set_caption("Game")

PLATFORM_MIN_WIDTH = 20
PLATFORM_MAX_WIDTH = 50
PLATFORM_MIN_DISTANCE = 50
PLATFORM_MAX_DISTANCE = 130

SURVIVOR_COLOR = (40, 40, 255)  # blue
OFFSPRING_COLOR = (10, 255, 10)  # green
NEW_BLOOD_COLOR = (255, 20, 20)  # red

GENOME_LEN = 5  # DO NOT CHANGE


def generate_genome():
    genome = []
    genome.append(random.randint(50, 300))  # how high is willing to try to jump
    genome.append(random.randint(0, 200))
    genome.append(random.randint(0, 300))
    genome.append(random.randint(0, 100))
    genome.append(random.randint(-100, 100))

    assert len(genome) == GENOME_LEN
    return genome


class DebugLine(pygame.sprite.Sprite):
    def __init__(self, height):
        super().__init__()
        self.surf = pygame.Surface((WIDTH, 5))
        self.surf.fill((70, 70, 70))
        self.rect = self.surf.get_rect(center=(WIDTH / 2, height))


class InfPlayer(pygame.sprite.Sprite):
    def __init__(self, spawnpoint, genome):
        super(InfPlayer, self).__init__()
        self.spawn = spawnpoint
        self.surf = pygame.Surface((15, 20))
        self.surf.fill((255, 255, 255))
        self.rect = self.surf.get_rect(center=spawnpoint)
        self.rank = -1

        self.prev_pos = self.rect.copy()
        self.speed = 6
        self.momentum = {"up_down": 0, "sideways": 0}
        self.gravity = 2
        self.can_jump = False
        self.top_speed = self.speed * 2

        self.jump_timer = 0
        self.jump_duration = 500  # milliseconds
        self.max_jump_height = -20
        self.min_jump_height = -5

        assert len(genome) == GENOME_LEN
        self.genome = genome
        self.target_pl = None
        self.highest_touched = HEIGHT * 2  # smaller -- better  ..i know, bad naming and all
        self.origin = None

    def reset_self(self):
        self.rect.center = self.spawn
        self.rank = -1
        self.highest_touched = HEIGHT * 2
        self.target_pl = None

    def update(self, pressed_keys, platforms):
        self.prev_pos = self.rect.copy()
        self.rect.move_ip(0, self.gravity)
        self.rect.move_ip(self.momentum["sideways"], self.momentum["up_down"])

        # Update momentum
        self.momentum["up_down"] += 1  # Fall faster
        if self.momentum["sideways"] > 0:
            if self.momentum["sideways"] < 2:
                self.momentum["sideways"] = 0

            self.momentum["sideways"] -= 2
        elif self.momentum["sideways"] < 0:
            if self.momentum["sideways"] > -2:
                self.momentum["sideways"] = 0
            self.momentum["sideways"] += 2

        # Check for user input and move accordingly
        if K_UP in pressed_keys and self.can_jump:
            self.momentum["up_down"] = self.max_jump_height
            self.rect.move_ip(0, -self.speed)
        # Check for jump input
        if K_UP in pressed_keys and self.can_jump:
            if self.jump_timer == 0:  # just started jumping
                self.jump_timer = pygame.time.get_ticks()
            else:  # already jumping
                elapsed_time = pygame.time.get_ticks() - self.jump_timer
                if elapsed_time > self.jump_duration:
                    elapsed_time = self.jump_duration
                jump_progress = elapsed_time / self.jump_duration
                jump_height = int(self.min_jump_height + jump_progress * (self.max_jump_height - self.min_jump_height))
                self.momentum["up_down"] = jump_height

        # Check for end of jump input
        if not K_UP in pressed_keys and self.jump_timer > 0:
            self.jump_timer = 0
            if self.momentum["up_down"] < -10:  # cap minimum jump height
                self.momentum["up_down"] = -10

        if K_DOWN in pressed_keys:
            self.rect.move_ip(0, self.speed)
        if K_LEFT in pressed_keys:
            self.rect.move_ip(-self.speed, 0)
            self.momentum["sideways"] -= self.speed // 2
        if K_RIGHT in pressed_keys:
            self.rect.move_ip(self.speed, 0)
            self.momentum["sideways"] += self.speed // 2

        # Limit momentum
        if self.momentum["sideways"] > self.top_speed:
            self.momentum["sideways"] = self.top_speed
        if self.momentum["sideways"] < -self.top_speed:
            self.momentum["sideways"] = -self.top_speed
        if self.momentum["up_down"] > self.top_speed:
            self.momentum["up_down"] = self.top_speed

        self.platform_collision_fix(platforms)

    def platform_collision_fix(self, platforms):
        # Check if player have collided with any Platform
        collided_platforms = pygame.sprite.spritecollide(self, platforms, False)

        if not collided_platforms:
            self.can_jump = False

        for platform in collided_platforms:
            if self.rect.bottom > platform.rect.top:  # touched from top
                if self.prev_pos.bottom >= platform.rect.top:  # came from top
                    self.rect.bottom = platform.rect.top
                    self.can_jump = True
                    self.momentum["up_down"] = 0

    def death(self, rank):
        self.rank = rank

    def next_step(self, platforms):
        """
        here is the "brain" of each player
        :param platforms:
        :return:
        """
        # self.can_jump = True
        # self.momentum = {"up_down": 0, "sideways": 0}
        # self.jump_timer = 0
        # self.genome  # list of floats?

        A = self.genome[0]  # max jump height
        B = self.genome[1]  # far horizontally
        C = self.genome[2]  # when falling, how far should be "saving" platform
        D = self.genome[3]  # how close to edge of platform is player willing to go
        E = self.genome[4]  # how far above target_pl to stop using K_UP

        if random.random() < 0.05:
            pressed = []
            return pressed
        # choosing which platform to jump to:
        if self.can_jump:
            # self.surf.fill((255, 0, 255))
            if not self.target_pl:  # does not have a target
                self.target_pl = misc.get_highest_sprite(self.rect.bottom, self.rect.bottom - A, platforms)
                if not self.target_pl:
                    self.surf.fill((100, 0, 0))  # remove later
                    # line = DebugLine(self.rect.bottom - A)  # remove later
                    # all_sprites.add(line)  # remove later

                    pressed = []
                    return pressed
            elif self.target_pl.rect.top < self.rect.bottom - A:  # does not believe he can jump that high
                self.target_pl = misc.get_highest_sprite(self.rect.bottom, self.rect.bottom - A, platforms)
                if not self.target_pl:
                    pressed = []
                    return pressed

            elif self.target_pl.rect.top >= self.rect.bottom:  # target is lower than player (and player can jump)
                # self.target_pl.surf.fill((0, 255, 0))  # return platform default color
                # old_target = self.target_pl
                # platforms.remove(old_target)
                self.target_pl = misc.get_highest_sprite(self.rect.bottom, self.rect.bottom - A, platforms)
                if not self.target_pl:
                    pressed = []
                    return pressed
                # platforms.add(old_target)

            # self.target_pl.surf.fill((50, 5, 200))  # target platform is blue

            standing_pl = misc.find_sprite_standing_on(platforms, self.rect.bottom)  # platform player stands on
            if self.highest_touched > standing_pl.rect.top:  # update for score
                self.highest_touched = standing_pl.rect.top
            standing_pl.surf.fill((200, 5, 5))  # standing platform is red
            # they are far away on horizontal axis
            if abs(self.target_pl.rect.center[0] - self.rect.center[0]) > B:  # centers are further away than B
                if self.target_pl.rect.center[0] > self.rect.center[0]:
                    if standing_pl.rect.right > self.rect.center[0] + D:  # how close to an end of platform
                        pressed = [K_RIGHT]
                        return pressed
                    else:
                        pressed = [K_UP, K_RIGHT]
                        return pressed
                else:
                    if standing_pl.rect.left < self.rect.center[0] - D:
                        pressed = [K_LEFT]
                        return pressed
                    else:
                        pressed = [K_UP, K_LEFT]
                        return pressed
            else:
                pressed = [K_UP]
                return pressed
        else:
            # self.surf.fill((0, 255, 255))
            if not self.target_pl or (
                    self.momentum["up_down"] > 0 and self.target_pl.rect.top < self.rect.bottom):  # falling down
                # find platform to fall to
                self.target_pl = misc.get_highest_sprite(self.rect.bottom + C, self.rect.bottom, platforms)

                if not self.target_pl:
                    pressed = []
                    return pressed
                # self.target_pl.surf.fill((255, 255, 0))
                return []
            elif self.target_pl.rect.top > self.rect.bottom - E:  # still going up and above target
                # above target platform
                if self.target_pl.rect.left < self.rect.center[0] < self.target_pl.rect.right:
                    pressed = []
                    return pressed
                elif self.target_pl.rect.right > self.rect.center[0]:
                    pressed = [K_UP, K_RIGHT]
                    return pressed
                elif self.target_pl.rect.left < self.rect.center[0]:
                    pressed = [K_UP, K_LEFT]
                    return pressed
            else:  # still going up and below target
                # (self.target_pl.rect.top, self.rect.bottom)
                # below target platform
                if self.target_pl.rect.left < self.rect.center[0] < self.target_pl.rect.right:
                    pressed = [K_UP]
                    return pressed
                elif self.target_pl.rect.right > self.rect.center[0]:
                    pressed = [K_UP, K_RIGHT]
                    return pressed
                elif self.target_pl.rect.left < self.rect.center[0]:
                    pressed = [K_UP, K_LEFT]
                    return pressed

        print("++++++++++++++++++++++++++++++++++++++random")
        return misc.decode_action(random.randint(0, 5))


def mutate(player: InfPlayer, severity: float):
    chromosome = random.randint(0, GENOME_LEN - 1)
    player.genome[chromosome] += player.genome[chromosome] * random.uniform(-severity, severity)


class InfAI:
    def __init__(self):
        self.spawnpoint = (WIDTH / 2, HEIGHT - 70)
        self.population = []  # all players
        self.players = []  # players that still play
        self.first_it = True
        self.survivors = 5
        self.new_blood = 5
        self.size = self.survivors + self.new_blood + 5
        self.tournament_size = 5
        self.mut_rate = 0.07
        self.mut_severity = 0.2

    def prepare(self, all_sprites):
        if self.first_it:
            self.first_it = False
            while len(self.population) < self.size:
                genome = generate_genome()
                P = InfPlayer(self.spawnpoint, genome)
                self.players.append(P)
                self.population.append(P)
                all_sprites.add(P)
        else:
            self.create_next_generation(all_sprites)

    """
    # unnecessary
    def end_of_iteration(self):
        # update so it takes into account highest touched platform
        sorted_players = sorted(sorted(self.population, key=lambda x: x.rank), key=lambda x: x.highest_touched)
        # print([p.highest_touched for p in sorted_players])
        # print("best player's genome:", sorted_players[0].genome)
    """

    def get_best(self):
        # update so it takes into account highest touched platform
        sorted_players = sorted(sorted(self.population, key=lambda x: x.rank),
                                key=lambda x: x.highest_touched)  # sorts smallest first, which works in this case
        return sorted_players[0:3]  # i know i don't need to sort them...

    def select_parents(self):
        """Selects two parents from the population using tournament selection."""
        tournament = random.sample(self.population, self.tournament_size)
        winner = min(tournament, key=lambda x: x.highest_touched)
        tournament.remove(winner)
        runner_up = min(tournament, key=lambda x: x.highest_touched)
        return winner, runner_up

    def crossover(self, parent1: InfPlayer, parent2: InfPlayer):
        """Performs random combination between two parents."""
        child = InfPlayer(self.spawnpoint,
                          genome=[random.choice(pair) for pair in zip(parent1.genome, parent2.genome)])
        return child

    def create_next_generation(self, all_sprites):
        """Creates the next generation of the population."""
        new_population = []
        # select survivors from old generation
        survivors = sorted(sorted(self.population, key=lambda x: x.highest_touched), key=lambda x: x.rank)[
                    :self.survivors]
        for survivor in survivors:
            # reset survivors
            survivor.reset_self()
            survivor.surf.fill(SURVIVOR_COLOR)
            survivor.origin = "survivor"
            all_sprites.remove(survivor)
            new_population.append(survivor)

        for _ in range(self.new_blood):
            genome = generate_genome()
            newbie = InfPlayer(self.spawnpoint, genome)
            newbie.surf.fill(NEW_BLOOD_COLOR)
            newbie.origin = "new_blood"
            # newbie.can_jump = False
            new_population.append(newbie)

        while len(new_population) < self.size:
            parent1, parent2 = self.select_parents()
            offspring = self.crossover(parent1, parent2)
            offspring.surf.fill(OFFSPRING_COLOR)
            offspring.origin = "offspring"

            if random.random() < self.mut_rate:
                mutate(offspring, self.mut_severity)
            new_population.append(offspring)

        self.players = []
        self.population = []
        for p in new_population:
            self.players.append(p)
            self.population.append(p)
            all_sprites.add(p)
        return new_population

    def next_step(self, platforms):
        fallen = set()
        for P in self.players:
            pressed = P.next_step(platforms)
            P.update(pressed, platforms)
            if P.rect.top > HEIGHT or P.rect.left < -10 or P.rect.right > WIDTH + 10:  # this works
                fallen.add(P)
                P.death(self.size - len(self.players))
                P.kill()
                # running = False

        self.players = [x for x in self.players if x not in fallen]
        if not self.players:
            return False, None

        highest_player = min(self.players, key=lambda x: x.rect.center[1])
        return True, highest_player


class Platform(pygame.sprite.Sprite):
    def __init__(self, center, size=None):
        super().__init__()
        if size:
            self.surf = pygame.Surface(size)
        else:
            self.surf = pygame.Surface((random.randint(PLATFORM_MIN_WIDTH, PLATFORM_MAX_WIDTH), 12))
        self.surf.fill((0, 255, 0))
        self.rect = self.surf.get_rect(center=center)


class InfLevel:
    def __init__(self, AI):
        pygame.init()

        self.platforms = pygame.sprite.Group()
        self.all_sprites = pygame.sprite.Group()

        # Set up the clock for controlling framerate
        self.clock = pygame.time.Clock()
        # Create the screen object
        self.screen = pygame.display.set_mode((WIDTH, HEIGHT))
        self.running = True

        self.height_reached = 0
        self.AI = AI

    def reset(self):
        self.height_reached = 0
        self.platforms.empty()
        self.all_sprites.empty()
        # self.AI.end_of_iteration()

    def plat_gen(self):
        while len(self.platforms) < 18:
            highest_pl = misc.find_highest_sprite(self.platforms)
            diff = random.randint(PLATFORM_MIN_DISTANCE, PLATFORM_MAX_DISTANCE)
            new_val = highest_pl.rect.top - diff

            # hi
            p = Platform((random.randint(30, WIDTH - 30), new_val))

            if p.rect.left < 0:
                p.rect.left = 0
            if p.rect.right > WIDTH:
                p.rect.right = WIDTH

            self.platforms.add(p)
            self.all_sprites.add(p)

    def run_once(self):

        PT1 = Platform((WIDTH / 2, HEIGHT - 20), size=(WIDTH, 20))
        self.AI.prepare(self.all_sprites)

        PT1.surf.fill((200, 50, 50))

        PT1.rect.left = 0
        self.all_sprites.add(PT1)
        self.platforms.add(PT1)

        """
        for i in range(5, 2):
            pl = Platform((WIDTH / 2, HEIGHT - 20), size=(WIDTH, 12+i))
            self.platforms.add(pl)
            self.all_sprites.add(pl)
        """

        plts = generate_sorted_list(13, PT1.rect.top - 30)
        for height in plts:
            center = (random.randint(0, WIDTH), height)
            pl = Platform(center)
            if pl.rect.left < 0:
                pl.rect.left = 0
            elif pl.rect.right > WIDTH:
                pl.rect.right = WIDTH
            self.platforms.add(pl)
            self.all_sprites.add(pl)

        running = True
        timer = 0
        while running:
            move_up = 0
            # slowly move the map
            timer += 1
            if timer >= 3:
                timer = 0
                move_up += 1

            running, highest_player = AI.next_step(self.platforms)
            if not running:
                break
            # in case some player is too fast
            if highest_player.rect.top <= HEIGHT / 6:
                move_up += abs(highest_player.momentum['up_down'])

            # move everything up
            self.height_reached += move_up
            for sprite in self.platforms:
                sprite.rect.y += move_up
                if type(sprite) == InfPlayer:
                    sprite.highest_touched += move_up

            for P in AI.population:
                P.rect.y += move_up
                P.highest_touched += move_up

            for plat in self.platforms:
                if plat.rect.top >= HEIGHT:
                    plat.kill()

            for event in pygame.event.get():
                if event.type == QUIT:
                    pygame.quit()
                    sys.exit()

            # generate new platforms
            self.plat_gen()
            displaysurface.fill((0, 0, 0))

            for entity in self.all_sprites:
                displaysurface.blit(entity.surf, entity.rect)

            pygame.display.update()
            FramePerSec.tick(FPS)

        self.platforms.empty()
        self.all_sprites.empty()


def generate_sorted_list(length, max_val):
    result = [random.randint(max_val - PLATFORM_MIN_DISTANCE, max_val)]
    for _ in range(length - 1):
        diff = random.randint(PLATFORM_MIN_DISTANCE, PLATFORM_MAX_DISTANCE)
        new_val = result[-1] - diff

        result.append(new_val)
    return result


def plot_heights(heights, categories=None):
    # Generate x-axis values (assuming 0-based index)
    x_values = range(len(heights))

    # Plot the graph
    if categories:
        # Create a dictionary to map categories to numeric values
        category_mapping = {'survivor': 'blue', 'offspring': 'green', 'new_blood': 'red', None: 'black'}

        # Map the categories to numeric values
        numeric_categories = [category_mapping[category] for category in categories]

        plt.scatter(x_values, heights, c=numeric_categories)

        # Create a legend with the corresponding categories
        legend_elements = [plt.Line2D([0], [0], marker='o', color='w', label=category,
                                      markerfacecolor=color, markersize=10)
                           for category, color in category_mapping.items()]

        # Add legend
        plt.legend(handles=legend_elements)

    else:
        plt.scatter(x_values, heights)

    # Set x-axis and y-axis labels
    plt.xlabel('Iteration')
    plt.ylabel('Height')

    # Set the title of the graph
    plt.title('Heights Reached')

    # Show the graph
    plt.show()


def describe_genomes(genomes):
    print(pd.DataFrame(np.array(genomes)).describe())


def plot_similarities(genomes):
    # Plot the similarity matrix
    plt.imshow(cosine_similarity(np.array(genomes)), cmap='hot', interpolation='nearest', vmin=0, vmax=1)

    # Add colorbar
    plt.colorbar()

    # Set x-axis and y-axis labels
    # plt.xlabel('Items')
    # plt.ylabel('Items')

    plt.title('Genome Similarities between runs')

    # Show the plot
    plt.show()


def print_genomes(data, per_line=4):
    print("[")
    for idx, item in enumerate(data):
        print(item, end=', ')

        if (idx + 1) % per_line == 0 or idx == len(data) - 1:
            print()
    print("]")


if __name__ == "__main__":
    from sklearn.metrics.pairwise import cosine_similarity
    import pandas as pd

    # import time

    best_genomes = []
    for j in range(1, 8):
        # print(f"run {j:2d} started")
        AI = InfAI()
        level = InfLevel(AI)
        heights = []
        winners = []
        who_won = {"offspring": 0, "new_blood": 0, "survivor": 0, None: 0}
        for i in range(j * 30 + 1):
            # print('.', end='')
            level.reset()

            # start = time.perf_counter()
            level.run_once()
            # finish = time.perf_counter()

            # print(f" iteration {i} took: {finish-start} seconds", end=', ')
            # print(f"height reached: {level.height_reached}")
            heights.append(level.height_reached)
            best = AI.get_best()
            # print(f"{level.height_reached} 3 best got to heights: {best[0].highest_touched}, "
            #       f"{best[1].highest_touched}, {best[2].highest_touched}")

            winners.append(best[0].origin)
            who_won[best[0].origin] += 1

        print(who_won)
        print(f"offspring: {(who_won['offspring'] / (j * 30)):.2f}")
        print(f"survivors: {(who_won['survivor'] / (j * 30)):.2f}")
        print(f"new_blood: {(who_won['new_blood'] / (j * 30)):.2f}")
        # print()
        best = AI.get_best()[0]
        best_genomes.append(best.genome)

        plot_heights(heights, winners)

    # print_genomes(best_genomes, 10)

    describe_genomes(best_genomes)
    # plot_similarities(best_genomes)
