import os
import pickle
from typing import Tuple

from constants import LEVEL_PATH, K_LEFT, K_UP, K_RIGHT, COLOUR_CHANGE


class Playground:
    def __init__(self, platforms=None, start=None, checkpoints=None, finish=None):
        """
        position = (x,y), size = (width,height)
        :param platforms: positions and sizes of platforms
        :param start: position of start
        :param checkpoints: list of positions of checkpoints
        :param finish: position of finish
        """
        if checkpoints is None:
            checkpoints = []
        self.platforms = platforms
        self.start = start
        self.checkpoints = checkpoints
        self.finish = finish


def prepare_save_map(special_points, platforms):
    pl = []
    checkpoints = []
    start = None
    finish = None
    for platform in platforms:
        # save tuples: topleft_pos + size
        pl.append((platform.rect.topleft, platform.rect.size))

    for spec in special_points:
        # save tuples: center_pos + action (checkpoint order should be fine)
        if spec.type == "Checkpoint":
            checkpoints.append(spec.rect.center)
        elif spec.type == "Start":
            start = spec.rect.center
        elif spec.type == "Finish":
            finish = spec.rect.center

    assert start
    assert finish
    return Playground(pl, start, checkpoints, finish)


def save_map(playground: Playground, level_name) -> None:
    pickle.dump(playground, open(LEVEL_PATH + level_name, "wb"))


def load_map() -> Tuple[Playground, str]:
    # Makes sure the levels path exists
    if os.path.exists(LEVEL_PATH):
        # Lists the maps in the path
        for item in os.listdir(LEVEL_PATH):
            print(item)

        # Asks for the name
        level = input("Enter a file name of the above: ")

        while not os.path.exists(LEVEL_PATH + level):
            # Asks for the name
            level = input("Enter a file name of the above: ")

        return pickle.load(open(LEVEL_PATH + level, "rb")), level


def load_map_alt(map_name):
    return pickle.load(open(LEVEL_PATH + map_name, "rb"))


def distance(point1, point2):
    x1, y1 = point1
    x2, y2 = point2
    return ((x2 - x1) ** 2 + (y2 - y1) ** 2) ** 0.5


def get_numeric_input(default=None, minimum=None, maximum=None, text='', num_type='int'):
    while True:
        inp = input(f"{text} (default {default}): ")
        if not inp:
            return default

        if num_type == 'int':
            if not inp.isdecimal():
                print("enter integer")
                continue
            num_inp = int(inp)
        else:
            assert num_type == 'float'
            num_inp = float(inp)

        if not minimum <= num_inp or not num_inp <= maximum:
            print(f"invalid range, choose number from interval {min}-{max}")
            continue

        return num_inp


# action encoding:
# 0 - None
# 1 - K_LEFT
# 2 - K_LEFT + K_UP
# 3 - K_UP
# 4 - K_RIGHT + K_UP
# 5 - K_RIGHT
def encode_action(pressed):
    action = 6
    if not pressed:
        action = 0
    elif pressed == [K_LEFT]:
        action = 1
    elif pressed == [K_LEFT, K_UP]:
        action = 2
    elif pressed == [K_UP]:
        action = 3
    elif pressed == [K_UP, K_RIGHT]:
        action = 4
    elif pressed == [K_RIGHT]:
        action = 5
    return action


def decode_action(action):
    pressed = []
    if action == 0:
        pressed = []
    elif action == 1:
        pressed = [K_LEFT]
    elif action == 2:
        pressed = [K_LEFT, K_UP]
    elif action == 3:
        pressed = [K_UP]
    elif action == 4:
        pressed = [K_UP, K_RIGHT]
    elif action == 5:
        pressed = [K_RIGHT]
    return pressed


def change_color(old_color):
    """cyclically changes colour"""
    cur_color = old_color.copy()
    # Pour from bin1 to bin2
    if cur_color[0] > 0 and cur_color[2] == 0:
        poured = min(cur_color[0], COLOUR_CHANGE)
        cur_color[0] -= poured
        cur_color[1] += poured
    # Pour from bin2 to bin3
    elif cur_color[1] > 0 and cur_color[0] == 0:
        poured = min(cur_color[1], COLOUR_CHANGE)
        cur_color[1] -= poured
        cur_color[2] += poured
    # Pour from bin3 to bin1
    elif cur_color[2] > 0 and cur_color[1] == 0:
        poured = min(cur_color[2], COLOUR_CHANGE)
        cur_color[2] -= poured
        cur_color[0] += poured

    return cur_color


def find_lowest_sprite(group):
    lowest_sprite = None
    for sprite in group:
        if lowest_sprite is None or sprite.rect.bottom > lowest_sprite.rect.bottom:
            lowest_sprite = sprite
    return lowest_sprite


def find_highest_sprite(group):
    highest_sprite = None
    for sprite in group:
        if highest_sprite is None or sprite.rect.bottom < highest_sprite.rect.bottom:
            highest_sprite = sprite
    return highest_sprite


def find_sprite_standing_on(group, top):
    closest = None
    for sprite in group:
        if sprite.rect.top == top:
            return sprite

        if not closest:
            closest = sprite
            continue
        if sprite.rect.top > top and abs(sprite.rect.top - top) < abs(closest.rect.top - top):
            closest = sprite
    return closest


def get_highest_sprite(lower_bound, upper_bound, sprite_group):
    highest_sprite = None
    closest_sprite = None
    closest_distance = None

    for sprite in sprite_group:
        if upper_bound <= sprite.rect.top < lower_bound:
            if not highest_sprite or sprite.rect.top < highest_sprite.rect.top:
                highest_sprite = sprite
        elif sprite.rect.top < lower_bound:
            dist = lower_bound - sprite.rect.top
            if not closest_sprite or dist < closest_distance:
                closest_sprite = sprite
                closest_distance = dist

    if highest_sprite:
        # print("returning nce")
        return highest_sprite
    elif closest_sprite:
        # print("returning bade")
        return
        return closest_sprite
    else:
        # print(f"lowerBound: {lower_bound}, upperBound: {upper_bound}")
        for sprite in sprite_group:
            if upper_bound <= sprite.rect.top <= lower_bound:
                # print(sprite.rect.top)
                pass
