import os

import misc
import sprites as sp
import pygame
from constants import *
from evolutionary_alg import EvolAI
from ant_colony import AntColony
from ant_colony_original import AntColonyOrigin

FPS = 120


def choose_AI():
    # return EvolAI
    print("choose AI\n"
          "0 - Ant Colony\n"
          "1 - Evolution Algorithm\n"
          "2 - Original Ant Colony")
    choice = misc.get_numeric_input(0, 0, 2, '', 'int')
    if choice == 0:
        return AntColony
    elif choice == 1:
        return EvolAI
    elif choice == 2:
        return AntColonyOrigin
    return AntColony


class Level:
    def __init__(self, playground: misc.Playground):
        pygame.init()

        self.playground = playground
        self.spawnpoint = playground.start
        self.platforms = pygame.sprite.Group()
        self.poi = pygame.sprite.Group()
        self.all_sprites = pygame.sprite.Group()
        self.poi_list = []
        self.next_poi_idx = 0

        # Set up the clock for a decent framerate
        self.clock = pygame.time.Clock()
        # Create the screen object
        # The size is determined by the constant SCREEN_WIDTH and SCREEN_HEIGHT
        self.screen = pygame.display.set_mode((SCREEN_WIDTH, SCREEN_HEIGHT))
        self.running = True

        self.reset()

    def reset(self):
        # Create groups to hold sprites
        # - platforms is used for collision detection and position updates
        # - poi == spec_idx-of-interest used for checkpoints and finish
        # - all_sprites is used for rendering
        self.platforms.empty()
        self.poi.empty()
        self.all_sprites.empty()

        self.poi_list = []
        self.next_poi_idx = 0

        for platform in self.playground.platforms:
            new_platform = sp.Platform(platform[0], platform[1])
            self.platforms.add(new_platform)
            self.all_sprites.add(new_platform)

        for checkpoint in self.playground.checkpoints:
            new_checkpoint = sp.PointOfInterest("Checkpoint", checkpoint)
            self.poi.add(new_checkpoint)
            self.all_sprites.add(new_checkpoint)
            self.poi_list.append(new_checkpoint)

        finish = sp.PointOfInterest("Finish", self.playground.finish)
        self.poi.add(finish)
        self.poi_list.append(finish)
        self.all_sprites.add(finish)

    def run_once(self, AI):
        # Main loop
        running = True
        while running and not AI.done:
            # for loop through the event queue
            for event in pygame.event.get():
                # Check for KEYDOWN event
                if event.type == KEYDOWN:
                    # If the Esc key is pressed, then exit the main loop
                    if event.key == K_ESCAPE:
                        running = False
                        # input()
                        break
                        # return False
                # Check for QUIT event. If QUIT, then set running to false.
                elif event.type == QUIT:
                    running = False
                    break
                    # return False

            # Fill the screen with black
            self.screen.fill((0, 0, 0))

            # Make next AI step
            AI.next_step()

            # Draw all sprites
            for entity in self.all_sprites:
                self.screen.blit(entity.surf, entity.rect)

            # Update the display
            pygame.display.flip()

            # Ensure program maintains a rate of at most 120 frames per second
            self.clock.tick(FPS)  # fps

        # if user quited, return False
        # (running == not AI.done)
        return running


def user_interface():
    # ask for level
    print("Choose map")
    # # load level
    playground, level_name = misc.load_map()
    level = Level(playground)
    # ask for AI
    # focus on entering parameters
    os.system('wmctrl -a simulation.py')

    print("Choose AI")
    AI_class = choose_AI()
    # # prepare AI (all must have same interface)
    AI = AI_class(level.spawnpoint, level.poi_list, level.platforms, level.all_sprites)
    # # (optionally) ask for parameters (inside "prepare" method)
    AI.prepare()

    max_iterations = misc.get_numeric_input(100, 1, 10000, "number of iterations")

    # focus on pygame window
    os.system('wmctrl -a pygame window')

    """
    # wait for keypress to start
    p = False
    while not p:
        # for loop through the event queue
        for event in pygame.event.get():
            # Check for KEYDOWN event
            if event.type == KEYDOWN:
                if event.key == K_ESCAPE:
                    p = True
    """
    # run
    first_finish = None
    first_path = None
    i = 0
    while i < max_iterations:
        i += 1
        print(f"iteration {i}:")
        if not level.run_once(AI):  # or AI.found_path:
            break  # user wants to quit
        if not first_finish and AI.found_path:
            first_finish = i
            first_path = AI.path
        AI.update_after_iteration()
        AI.reset_positions()
        # level.reset()

    if first_finish:
        print(f"first found path in iteration {first_finish}")
        print(f"first path has length: {len(first_path)}")
    if AI.ever_found_path:
        print(f"best found path has length: {len(AI.best_found_path)}")
    else:
        print("AI was not able to find path")


def tests():
    max_it = 60
    a = 0
    for mapname in ["e1", "e2", "e3", "m1", "m2", "m3", "h1", "h2", "h3", "vertical", "vertical2", "vertical3", "hole"]:
        if a % 3 == 0:
            max_it += 20
        a += 1
        print()
        for AI_class in [EvolAI, AntColony, AntColonyOrigin]:  # EvolAI, AntColony,
            # print("a", a)
            # a += 1
            playground = misc.load_map_alt(mapname)
            level = Level(playground)
            AI = AI_class(level.spawnpoint, level.poi_list, level.platforms, level.all_sprites)
            AI.prepare(use_default=True)

            first_finish = None
            first_path = None
            i = 0
            while i < max_it:
                i += 1
                # print(f"iteration {i}:")
                if not level.run_once(AI):  # or AI.found_path:
                    break  # user wants to quit
                if not first_finish and AI.found_path:
                    first_finish = i
                    first_path = AI.path
                AI.update_after_iteration()
                AI.reset_positions()
                # level.reset()

            print(f"AI: {AI_class.name:15s}, map:{mapname}, max_it: {max_it}, results --", end='')
            if first_finish:
                print(f" first path in it: {first_finish},"
                      f" first path len: {len(first_path)},", end='')
            if AI.ever_found_path:
                print(f" best path len: {len(AI.best_found_path)}")
            else:
                print(" AI was not able to find path")

            # write result to file
            with open('comparison.txt', 'a') as fd:
                if first_finish:
                    fd.write(f"AI: {AI_class.name:15s}, map:{mapname}, max_it: {max_it}, results --"
                             f" first path in it: {first_finish},"
                             f" first path len: {len(first_path)},"
                             f" best path len: {len(AI.best_found_path)}\n")
                else:
                    fd.write(f"AI: {AI_class.name:15s}, map:{mapname}, max_it: {max_it}, results --"
                             f" AI was not able to find path\n")

            # del level
            # del AI
            # del playground


if __name__ == "__main__":
    user_interface()
    # tests()
