import pygame
from constants import *
import random


# Define a player object by extending pygame.sprite.Sprite
# The surface drawn on the screen is now an attribute of 'player'
class Player(pygame.sprite.Sprite):
    def __init__(self, spawnpoint):
        super(Player, self).__init__()
        self.surf = pygame.Surface((15, 20))
        self.surf.fill((255, 255, 255))
        self.rect = self.surf.get_rect(center=spawnpoint)

        self.prev_pos = self.rect.copy()
        self.speed = 6
        self.momentum = {"up_down": 0, "sideways": 0}
        self.gravity = 2
        self.can_jump = True
        self.top_speed = self.speed * 2

        self.jump_timer = 0
        self.jump_duration = 500  # milliseconds
        self.max_jump_height = -20
        self.min_jump_height = -5

    def stop_momentum(self):
        self.momentum["up_down"] = 0
        self.momentum["sideways"] = 0

    def update(self, pressed_keys, platforms, spec_points):
        self.prev_pos = self.rect.copy()
        self.rect.move_ip(0, self.gravity)
        self.rect.move_ip(self.momentum["sideways"], self.momentum["up_down"])

        # Update momentum
        self.momentum["up_down"] += 1  # Fall faster
        if self.momentum["sideways"] > 0:
            if self.momentum["sideways"] < 2:
                self.momentum["sideways"] = 0

            self.momentum["sideways"] -= 2
        elif self.momentum["sideways"] < 0:
            if self.momentum["sideways"] > -2:
                self.momentum["sideways"] = 0
            self.momentum["sideways"] += 2

        # Check for user input and move accordingly
        if pressed_keys[K_UP] and self.can_jump:
            self.momentum["up_down"] = self.max_jump_height
            self.rect.move_ip(0, -self.speed)
        # Check for jump input
        if pressed_keys[K_UP] and self.can_jump:
            if self.jump_timer == 0:  # just started jumping
                self.jump_timer = pygame.time.get_ticks()
            else:  # already jumping
                elapsed_time = pygame.time.get_ticks() - self.jump_timer
                if elapsed_time > self.jump_duration:
                    elapsed_time = self.jump_duration
                jump_progress = elapsed_time / self.jump_duration
                jump_height = self.min_jump_height + jump_progress * (self.max_jump_height - self.min_jump_height)
                self.momentum["up_down"] = jump_height

        # Check for end of jump input
        if not pressed_keys[K_UP] and self.jump_timer > 0:
            self.jump_timer = 0
            if self.momentum["up_down"] < -10:  # cap minimum jump height
                self.momentum["up_down"] = -10

        if pressed_keys[K_DOWN]:
            self.rect.move_ip(0, self.speed)
        if pressed_keys[K_LEFT]:
            self.rect.move_ip(-self.speed, 0)
            self.momentum["sideways"] -= self.speed // 2
        if pressed_keys[K_RIGHT]:
            self.rect.move_ip(self.speed, 0)
            self.momentum["sideways"] += self.speed // 2

        # Limit momentum
        if self.momentum["sideways"] > self.top_speed:
            self.momentum["sideways"] = self.top_speed
        if self.momentum["sideways"] < -self.top_speed:
            self.momentum["sideways"] = -self.top_speed
        if self.momentum["up_down"] > self.top_speed:
            self.momentum["up_down"] = self.top_speed

        # Keep player on the screen
        if self.rect.left < 0:
            self.rect.left = 0
            self.stop_momentum()
        if self.rect.right > SCREEN_WIDTH:
            self.rect.right = SCREEN_WIDTH
            self.stop_momentum()
        if self.rect.top <= 0:
            self.rect.top = 0
            self.stop_momentum()
        if self.rect.bottom >= SCREEN_HEIGHT:
            self.rect.bottom = SCREEN_HEIGHT
            self.stop_momentum()

        self.platform_collision_fix(platforms)

    def platform_collision_fix(self, platforms):
        # Check if player have collided with any Platform
        collided_platforms = pygame.sprite.spritecollide(self, platforms, False)

        if not collided_platforms:
            self.can_jump = False

        for platform in collided_platforms:
            self.can_jump = True

            if self.rect.bottom > platform.rect.top:  # touched from top
                if self.prev_pos.bottom > platform.rect.top:  # came from side
                    if self.prev_pos.right <= platform.rect.left:  # came from right
                        self.rect.right = platform.rect.left
                        self.momentum["sideways"] = 0
                        self.momentum["up_down"] //= 2
                    elif self.prev_pos.left >= platform.rect.right:  # came from left
                        self.rect.left = platform.rect.right
                        self.momentum["sideways"] = 0
                        self.momentum["up_down"] //= 2
                    else:  # came from bottom
                        self.rect.top = platform.rect.bottom
                        self.momentum["up_down"] = 0
                        self.can_jump = False
                else:
                    self.rect.bottom = platform.rect.top
                    self.momentum["up_down"] = 0


class Platform(pygame.sprite.Sprite):
    def __init__(self, position, size, color=(255, 255, 255)):
        """

        :param position: tuple (x_pos, y_pos)
        :param size: tuple (width, height)
        """
        super(Platform, self).__init__()
        self.surf = pygame.Surface(size)
        self.color = color
        self.surf.fill(color)
        self.rect = self.surf.get_rect(topleft=position)

    def resize(self, new_size, new_position):
        self.surf = pygame.transform.scale(self.surf, new_size)
        self.surf.fill(self.color)
        self.rect = self.surf.get_rect(topleft=new_position)


class SpecialPoints(pygame.sprite.Sprite):
    checkpoint_number = 0
    start = None
    finish = None
    checkpoints = pygame.sprite.Group()

    def __init__(self, type, position):
        width = 20
        height = 25
        yellow = (255, 255, 0)
        super(SpecialPoints, self).__init__()
        self.type = type
        self.surf = pygame.Surface((width, height))
        self.rect = self.surf.get_rect(center=position)

        if type == "Checkpoint":
            SpecialPoints.checkpoint_number += 1
            font = pygame.font.SysFont("Arial", 18)
            self.surf.fill((100, 100, 0))
            self.textSurf = font.render(f"{SpecialPoints.checkpoint_number}", True, yellow)
            W = self.textSurf.get_width()
            H = self.textSurf.get_height()
            self.surf.blit(self.textSurf, [width / 2 - W / 2, height / 2 - H / 2])
            SpecialPoints.checkpoints.add(self)

        elif type == "Start":
            if SpecialPoints.start:
                SpecialPoints.start.kill()
            self.surf.fill((0, 230, 0))
            SpecialPoints.start = self

        elif type == "Finish":
            if SpecialPoints.finish:
                SpecialPoints.finish.kill()
            self.surf.fill((230, 0, 0))
            SpecialPoints.finish = self
        else:
            assert False
        self.rect = self.surf.get_rect(center=position)

    def kill(self):
        if self.type == "Checkpoint":
            SpecialPoints.checkpoint_number = 0
            for sp in SpecialPoints.checkpoints:
                sp.sup_kill()
        else:
            super().kill()

    def sup_kill(self):
        super().kill()

class PointOfInterest(pygame.sprite.Sprite):
    checkpoint_number = 0
    start = None
    finish = None

    def __init__(self, type, position):
        width = 20
        height = 25
        yellow = (255, 255, 0)
        super(PointOfInterest, self).__init__()
        self.type = type
        self.surf = pygame.Surface((width, height))
        self.rect = self.surf.get_rect(center=position)

        if type == "Checkpoint":
            SpecialPoints.checkpoint_number += 1
            font = pygame.font.SysFont("Arial", 18)
            self.surf.fill((100, 100, 0))
            self.textSurf = font.render(f"{SpecialPoints.checkpoint_number}", True, yellow)
            W = self.textSurf.get_width()
            H = self.textSurf.get_height()
            self.surf.blit(self.textSurf, [width / 2 - W / 2, height / 2 - H / 2])
            SpecialPoints.checkpoints.add(self)

        elif type == "Start":
            if SpecialPoints.start:
                SpecialPoints.start.kill()
            self.surf.fill((0, 230, 0))
            SpecialPoints.start = self

        elif type == "Finish":
            if SpecialPoints.finish:
                SpecialPoints.finish.kill()
            self.surf.fill((230, 0, 0))
            SpecialPoints.finish = self
        else:
            assert False
        self.rect = self.surf.get_rect(center=position)
